/*
 * Copyright (C) 2014 Yanjun Jiang <jiangy76@mcmaster.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the distribution
 *     * Neither the name of the McMaster Centre for Software Certification nor the names
 *       of its contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package ca.mcscert.jtet.eventbgenerator;

import ca.mcscert.jtet.expression.BooleanVariableType;
import ca.mcscert.jtet.expression.EventBGenerator;
import ca.mcscert.jtet.expression.VariableCollection;
import ca.mcscert.jtet.parsers.MatlabParser;
import ca.mcscert.jtet.tablularexpression.Cell;
import ca.mcscert.jtet.tablularexpression.HierarchcialGridCheckerGenerator;
import ca.mcscert.jtet.tablularexpression.TwoDimensionalGrid;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.List;

/**
 * This class generate common non-initialisation events
 * @author Yanjun Jiang <jiangy76@mcmaster.ca>
 */
final public class HierarchicalGridEventBGenerator implements HierarchcialGridCheckerGenerator {

    public HierarchicalGridEventBGenerator(VariableCollection variableDefinitions, String tableName, List<String> outputVariables, TwoDimensionalGrid outputGrid) {
        m_variableDefinitions = variableDefinitions;
        m_tableName = tableName;
        m_outputGrid = outputGrid;
        m_outputVariables = outputVariables;
    }

    private String getCurrentEventName(){
        return "Event" + m_currentEventNo;
    }

    private String generateEventLabelXml(int eventNo) {
        return "<org.eventb.core.event name=\"" + getCurrentEventName() + "\" org.eventb.core.convergence=\"0\" org.eventb.core.extended=\"false\" org.eventb.core.label=\"evt" + eventNo + "\">\n";
    }

    private String generateEventGuardXml(int guardNo, String guard) {
        return "<org.eventb.core.guard name=\"Guard" + guardNo + "\" org.eventb.core.label=\"grd" + + guardNo + "\" org.eventb.core.predicate=\"" + guard +"\"/>\n";
    }

    private String generateEventActionXml(int actionNo) {
        String value = m_outputGrid.get(0, m_currentEventNo - 1).contents();
        return "<org.eventb.core.action name=\"Action" + actionNo + "\" org.eventb.core.assignment=\"" + m_outputVariables.get(0) +  " ≔ " + value + "\" org.eventb.core.label=\"act" + actionNo + "\"/>\n";
    }

    private String generateEventActionsXml() {
        int actionNo = 1;
        String ret = "";

        ret += generateEventActionXml(actionNo);
        return ret;
    }

    private String generateEventXml(String currentCellContent) {
        int guardNo = 1;
        String ret = "";
        final String eventXmlPost = "</org.eventb.core.event>\n";

        ret += generateEventLabelXml(m_currentEventNo);
        for (String parentCellContent: m_parentCellsEventBCode){
            ret += generateEventGuardXml(guardNo++, parentCellContent);
        }

        ret += generateEventGuardXml(guardNo, currentCellContent);
        ret += generateEventActionsXml();
        ret += eventXmlPost;

        return ret;
    }

    private void outputCell(String currentCellContent) {
        m_eventsXml += generateEventXml(currentCellContent);
    }

    @Override
    public void descendIntoGridFromCell(Cell cell) {
        m_parentCellsEventBCode.addFirst(MatlabParser.parseMatlabCode(m_variableDefinitions, cell.contents()).getCheckerOutput(m_eventBGenerator, BooleanVariableType.Type));
    }

    @Override
    public void ascendFromGrid() {
        m_parentCellsEventBCode.removeFirst();
    }

    @Override
    public void handleEdgeCell(Cell cell) {
    }

    @Override
    public void handleLeafCell(Cell cell) {
        m_currentEventNo++;
        String parsedCellCode = handleCell(cell);
        outputCell(parsedCellCode);
    }

    public String handleCell(Cell cell) {
        return MatlabParser.parseMatlabCode(m_variableDefinitions, cell.contents()).getCheckerOutput(m_eventBGenerator, BooleanVariableType.Type);
    }

    @Override
    public String getFinalString() {
        return m_eventsXml;
    }

    int m_currentEventNo = 0;
    String m_eventsXml = "";
    String m_tableName = "";
    List<String> m_outputVariables;
    TwoDimensionalGrid m_outputGrid;
    VariableCollection m_variableDefinitions;
    Deque<String> m_parentCellsEventBCode = new ArrayDeque<String>();
    EventBGenerator m_eventBGenerator = new EventBGenerator();
}
